# $Id: Insertion_Sequence.pm,v 1.14 2007/02/26 23:27:53 cl Exp $

##
# The Insertion_Sequence class stores all the information about an Insertion
#	Sequence.
#


use strict;
use warnings;

use Bio::IScan::Utilities;

use Bio::AlignIO;
use Bio::LocatableSeq;
use Bio::Range;
use Bio::Root::Root;


package Bio::IScan::Insertion_Sequence;
	our @ISA = 'Bio::Root::Root';


##
# If the source_accession, source_locus, and source paramters are not provided
#	to the constructor, the constructor will attempt to determine them from the
#	the sequence parameter.
# If the display_id paramter is not provided to the constructor, a unique 
#	display_id will automatically be generated by concatenating the name,
#	source_locus, start, and length attributes using the following format:
#		<name>-<source-locus>-<start>-<length>.
#
sub new {
	my $class = shift;
	my $this = $class->SUPER::new();

	my $parameters = [
			'sequence', 
			'orfs', 
			'hsps', 
			'dr_alignment'
		];
	my $optional_parameters = {
			'name' => undef,
			'source_accession' => undef,
			'source_locus' => undef,
			'source' => undef,
			'display_id' => undef,
			'ir_alignment' => undef, 
			'l_ir_alignment' => undef, 
			'r_ir_alignment' => undef
		};

	my $arguments = Utilities::Process_Arguments($parameters, 
												 $optional_parameters,
												 @_);
	# Format is 'gb|accession|locus|source'
	my $id = Utilities::Parse_FastA_Header($arguments->{'sequence'}->display_id());
	if (!defined $arguments->{'source_accession'}) {
		$arguments->{'source_accession'} = $id->accession();
	}
	if (!defined $arguments->{'source_locus'}) {
		$arguments->{'source_locus'} = $id->locus();
	}
	if (!defined $arguments->{'source'}) {
		$arguments->{'source'} = $id->source();
	}

	if (!defined $arguments->{'ir_alignment'}) {
		if (!defined $arguments->{'l_ir_alignment'} || 
			!defined $arguments->{'r_ir_alignment'}) {

			$this->throw(-class => 'Bio::Root::BadParameter',
						 -text => "'ir_alignment' or 'l_ir_alignment' and 'r_ir_alignment' must be specified",
						 -value => "Invalid argument combination");
		}
	}
	# 'ir_alignment' is defined
	else {
		if (defined $arguments->{'l_ir_alignment'} && 
			defined $arguments->{'r_ir_alignment'}) {
			$this->throw(-class => 'Bio::Root::BadParameter',
						 -text => "'ir_alignment' or 'u_ir_alignment' and 'd_ir_alignment' must be specified",
						 -value => "Invalid argument combination");
		}
		$arguments->{'l_ir_alignment'} = $arguments->{'ir_alignment'};
		$arguments->{'r_ir_alignment'} = $arguments->{'ir_alignment'};
	}

	map { $this->{$_} = $arguments->{$_}; } keys %{ $arguments };

	return bless $this, $class;
}


##
# Gets or sets and returns the current display ID.
#
# @param [display_id]
#	The new display ID
#	
# @return
#	The current display ID (after any new display ID has been set).
#
sub display_id { # instance
	my $this = shift;
	
	if (@_) {
		$this->{'display_id'} = shift;
	}
	
	# If a display ID has not been generated or set, generate a display ID that
	#	will be unique for any ORF.	
	if (!defined $this->{'display_id'}) {
		$this->{'display_id'} = $this->name() . '-' . 
								$this->source_locus() . '-' . 
								$this->coding_start() . '-' .
								$this->coding_length();
	}	
	
	return $this->{'display_id'};
}


sub name { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'name'};
}

sub source_accession_number { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'source_accession'};
}

sub source_locus { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'source_locus'};
}

sub source { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'source'};
}

sub is_seq { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "IS sequence is immutable",
					 -value => @_);
	}
	
	return $this->{sequence};
}

sub start {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'sequence'}->start();
}

sub end {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'sequence'}->end();
}

sub length {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "length is immutable",
					 -value => @_);
	}
	
	return $this->end() - $this->start() + 1;
}

sub strand {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'sequence'}->strand();
}


## ORF methods

sub orf {
	my $this = shift;
	my $index = shift;
	
	my $orfs = $this->orfs();
	my $orf_count = scalar @{ $orfs };

	if ($index < 0 || $index > $orf_count - 1) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "ORF index out of range\n" .
							  "\tIndex: " . $index . "\n" .
							  "\tORF(s): " . $orf_count,
					 -value => $index);
	}
	
	return $orfs->[$index];
}

sub orfs { # instance
	my $this = shift;
	
	return $this->{'orfs'};
}

sub coding_start { # instance
	my $this = shift;

	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	my $coding_start = (sort({ 
		$a->start() <=> $b->start()
	} @{ $this->orfs() }))[0]->start();
	
	return $coding_start;
}

sub coding_end { # instance
	my $this = shift;

	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	my $coding_end = (sort({
		$b->end() <=> $a->end()
	} @{ $this->orfs() }))[0]->end();
	
	return $coding_end;
}

sub coding_length { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->coding_end() - $this->coding_start() + 1;
}

##
# Returns the combined coding lengths of all ORF(s), taking into account
#	overlapping ORF(s).
#
# @return
#	The combined coding lengths
#
# @throw Bio::Root::BadParameter
#	If any parameter is provided; the coding length is determined by the lengths
#		of the ORF(s) and the ORF distance, and can not be set.
#
# Note: This method is not safe with more than two ORFs.
#
sub combined_orf_length {
	my $this = shift;
	
	$this->throw(-class => 'Bio::Root::NotImplemented',
				 -text => "Method <combined_orf_length> is deprecated for maitenance."
				 -value => 'combined_orf_length');
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Attribute is immutable: coding_length",
					 -value => @_);
	}
	
	my $length = undef;
	
	$this->debug("IS ORFs: ", $#{ $this->{orfs} }, "\n");
	
	my $first_orf = $this->{orfs}->[0];
	my $last_orf = $this->{orfs}->[$#{ $this->{orfs} }];
	
	if (scalar @{ $this->{orfs} } > 1) {
		# If the ORFs overlap, the sequence of one ORF must be shortened before
		#   concatinating it to the sequence of the other ORF. It doesn't matter
		#   which ORF is shortened, so the left ORF is shortened.
		my $orf_distance = $this->orf_distance();
		$this->debug("ORF Distance: ", $orf_distance, "\n");
		
		if ($orf_distance < 0) {
			$length = $first_orf->length() + $orf_distance + $last_orf->length();
		}
		# The ORFs do not overlap
		else {
			$length = $first_orf->length() + $last_orf->length();
		}
	}
	# There is only 1 ORF
	else {
		$length = $this->{'orfs'}->[0]->length();
	}
	
	return $length;
}


##
# Note: This method is not safe with more than two ORFs.
#
sub coding_seq {
	my $this = shift;
	
	$this->throw(-class => 'Bio::Root::NotImplemented',
				 -text => "Method <coding_seq> is deprecated for maitenance."
				 -value => 'coding_seq');
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	$this->debug("IS ORFs: ", $#{ $this->{orfs} }, "\n");
	
	my $first_orf = $this->{orfs}->[0];
	my $last_orf = $this->{orfs}->[$#{ $this->{orfs} }];
	
	my $sequence = undef;
	
	if (scalar @{ $this->{orfs} } > 1) {
		# If the ORFs overlap, the sequence of one ORF must be shortened before
		#   concatinating it to the sequence of the other ORF. It doesn't matter
		#   which ORF is shortened, so the left ORF is shortened.
		my $orf_distance = $this->orf_distance();
		$this->debug("ORF Distance: ", $orf_distance, "\n");
		
		if ($orf_distance < 0) {
			my $left_end = $first_orf->length() + $orf_distance;
			$sequence = $first_orf->subseq(1, $left_end) . 
				$last_orf->seq();
		}
		# The ORFs do not overlap
		else {
			$sequence = $first_orf->seq() . 
			$last_orf->seq();
		}
	}
	# There is only 1 ORF
	else {
		$sequence = $first_orf->seq();
	}
	
	my $coding_seq = 
		Bio::LocatableSeq->new(-start => $first_orf->start(),
							   -end => $last_orf->end(),
							   -strand => $this->strand(),
							   -seq => $sequence);
	
	return $coding_seq;
}

sub orf_distance {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "orf_distance is immutable",
					 -value => @_);
	}
	
	my $distance = 0;
	
	if (scalar @{ $this->{hsp_pair} } > 1) {
		my $first_hsp = $this->{hsp_pair}->[0];
		my $last_hsp = $this->{hsp_pair}->[$#{ $this->{hsp_pair} }];
		$distance = $last_hsp->start('hit') - $first_hsp->end('hit');
	}
	
	return $distance;
}


sub hsps {
	my $this = shift;
	
	if (@_) {
		$this->{'hsps'} = shift;
	}
	
	return $this->{'hsps'};
}


## IR methods

sub ir_alignment {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'l_ir_alignment'};
}

sub left_ir_alignment {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'l_ir_alignment'};
}

sub right_ir_alignment {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'r_ir_alignment'};
}

sub left_ir_alignment_seq {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'l_ir_alignment'}->left_sequence();
}

sub right_ir_alignment_seq {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'r_ir_alignment'}->right_sequence();
}

sub left_ir_score {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'l_ir_alignment'}->score();
}

sub right_ir_score {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'r_ir_alignment'}->score();
}

sub ir_p_value { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	if ($this->{'l_ir_alignment'} != $this->{'r_ir_alignment'}) {
		return $this->{'l_ir_alignment'}->p_value() * 
			   $this->{'r_ir_alignment'}->p_value();
	}
	
	return $this->{'l_ir_alignment'}->p_value();
}


## DR methods

sub dr_alignment {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'dr_alignment'};
}

sub left_dr_alignment_seq {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}

	return $this->{'dr_alignment'}->left_sequence();
}

sub right_dr_alignment_seq {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'dr_alignment'}->right_sequence();
}

sub dr_score {
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Insertion_Sequence are immutable.",
					 -value => @_);
	}
	
	return $this->{'dr_alignment'}->score();
}


1;
