# $Id: Search_Response.pm,v 1.3 2006/11/29 07:09:38 cl Exp $


##
# A Search_Response packages all the response data produced by IScan when
#	searching for ISs.
#


use strict;

use Bio::Root::Root;


package Bio::IScan::Search_Response;
	our @ISA = 'Bio::Root::Root';


sub new { # constructor
	my $class = shift;
	my $this = $class->SUPER::new(@_);
	
	my $parameters = [];
	my $optional_parameters = ['search_request',
							   'version'];

	my $arguments = $this->_process_arguments($parameters, $optional_parameters, @_);
	map { $this->{$_} = $arguments->{$_} } keys %{ $arguments };
	
	$this->{insertion_sequences} = [];
	$this->{ir_statistics} = {};

	$this->{utc_begin} = time;
	
	return bless $this, $class;	
}


sub insertion_sequences { # instance
	my $this = shift;
	
	if (@_) {
		$this->_throw_immutable('insertion sequences');
	}
	
	return $this->{insertion_sequences};
}


sub search_request { # instance
	my $this = shift;
	
	if (@_) {
		$this->{search_request} = shift;
	}
	
	return $this->{search_request};
}


sub version { # instance
	my $this = shift;
	
	if (@_) {
		$this->_throw_immutable('version');
	}
	
	return $this->{version};
}


sub execution_begin { # instance
	my $this = shift;
	
	if (@_) {
		$this->_throw_immutable('execution_begin');
	}
	
	# In scalar context 'localtime' returns a time string
	return scalar localtime $this->{utc_begin};
}


sub mark_execution_end { # instance
	my $this = shift;
	
	$this->{utc_end} = 	time;
}


sub execution_end { # instance
	my $this = shift;
	
	if (@_) {
		$this->_throw_immutable('execution_end');
	}
	
	# In scalar context 'localtime' returns a time string
	return scalar localtime $this->{utc_end};
}


sub set_ir_statistics { # instance
	my $this = shift;

	my $sequence_id = shift;
	my $statistics = shift;
	
	if (!defined $sequence_id) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Argument not provided: sequence id",
					 -value => "$sequence_id");
	}
	
	if (!defined $statistics) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Argument not provided: statistics",
					 -value => "$sequence_id");
	}
	
	$this->{ir_statistics}->{$sequence_id} = $statistics;
}


sub ir_statistics { # instance
	my $this = shift;
	my $sequence_id = shift;
	my $statistics = shift;
	
	if (defined $statistics) {
		$this->{'ir_statistics'}->{$sequence_id} = $statistics;
	}

	return $this->{'ir_statistics'}->{$sequence_id};
}


sub add_insertion_sequence { # instance
	my $this = shift;
	
	push @{ $this->{insertion_sequences} }, shift;
}


sub _process_arguments { # instance
	my $this = shift;
	
	my $parameters = shift;
	my $optional_parameters = shift;
	my ($arguments) = @_;
	
	
	# Convert arguments to lowercase
	%{ $arguments } = map { lc($_) => $arguments->{$_} } keys %{ $arguments };
	
	# Make sure all the required parameters have been specified, and tranfer
	#   them over to class instance storage
	for my $test (@{ $parameters} ) {
		if (!defined $arguments->{$test}) {
			$this->throw(-class => 'Bio::Root::BadParameter',
						 -text => "Argument not specified: <$test>",
						 -value => $test);
		}
		
		$this->{$test} = $arguments->{$test};
	}
	
	for my $test (@{ $optional_parameters }) {
		if (defined $arguments->{$test}) {
			$this->{$test} = $arguments->{$test};
		}
	}
	
	return $arguments;
}


sub _throw_immutable { # instance
	my $this = shift;
	
	my $name = shift;
	
	$this->throw(-class => 'Bio::Root::BadParameter',
				 -text => "Value is immutable: ". $name,
				 -value => "the sequence_id");
}


@Search_Response_Exception::ISA = 'Bio::Root::Exception';


1;

__END__
