# $Id: Cache_Controller.pm,v 1.7 2007/02/26 23:09:05 cl Exp $


use strict;

use Bio::IScan::Global;
use Bio::IScan::Utilities;

use Bio::Root::Root;

use Storable;


package Bio::IScan::Server::Cache_Controller;
	our @ISA = 'Bio::Root::Root';


sub new {
	my $class = shift;
	my $this = $class->SUPER::new(@_);
	
	my $parameters = ['cache_path', 'cache_name'];
	my $optional_parameters = {};

	my $arguments = Utilities::Process_Arguments($parameters, 
												 $optional_parameters,
												 @_);
	map { $this->{$_} = $arguments->{$_} } keys %{ $arguments };

	$this->{persistent} = 1;
		
	if (!-e $this->{cache_path}) {
		$this->throw(-class => 'Statistics_Controller_Exception',
					 -text => "Cache path does not exist: " . $this->{cache_path},
					 -value => $!);
	}
	
	return bless $this, $class;	
}


##
# Returns an array of the files used for the statistics database this controller
#   is associated with.
#
# @return @strings
#   Array containing the filenames, with whatever path was specified in the
#		constructor (see @{new}), of the files associated with this statistics
#		database.
#
sub cache_file { # instance
	my $this = shift;
		
	my $filename =
		File::Spec->catfile($this->{cache_path}, $this->{cache_name});
		
	return $filename;	
}


sub cache { # instance
	my $this = shift;
	
	my $cache_data = shift;
		
	Storable::lock_nstore($cache_data, $this->cache_file());
}


sub cache_key { # instance
	my $this = shift;
	
	my $cache_data = shift;
	my $key = shift;

	$this->retrieve();

	$this->{cache}->{$key} = $cache_data;
	$this->cache($this->{cache});
}


sub retrieve { # instance
	my $this = shift;

	# If the cache file is not present this exception handling block will 
	#	prevent that exception from terminating the process.
	my $error = 0;	
	eval {
		if (!defined $this->{cache}) {
			$this->{cache} = Storable::lock_retrieve($this->cache_file());
		}
		$error = int $!;
	};	
	if ($@) {
		$this->{cache} = {};
	}
	
	return $this->{cache};
}


sub retrieve_key { # instance
	my $this = shift;
	
	my $key = shift;
	
	$this->retrieve();
	
	if (!exists $this->{cache}->{$key}) {
		$this->throw(-class => 'Cache_Controller_Exception',
					 -text => "Key does not exist: " . $key . "\n" .
							  "\tCache: " . $this->cache_file(),
					 -value => $key);
	}

	return $this->{cache}->{$key};
}


sub find_keys { # instance
	my $this = shift;
	
	my $search_key = shift;
	$this->_debug("Searching for keys matching: ", $search_key);
	
	$this->retrieve();
	
	my @cache_keys = keys %{ $this->{cache} };
	my @keys = grep /^$search_key/, @cache_keys;		

	for my $key (@keys) {
		$this->_debug("Found key: ", $key);
	}

	return @keys;
}	

	
sub generate_key { # instance
	my $this = shift;
	my $key_components = shift;

	my $key = "";
	
	for my $key_component (@{ $key_components }) {
		$key .= $key_component . ':';
	}

	return $key;
}


sub _debug { # instance
	my $this = shift;
	Global::Debug(__PACKAGE__, @_);
}


@Cache_Controller_Exception::ISA = 'Bio::Root::Exception';


1;

__END__
