# $Id: IScan.pm,v 1.35 2007/05/28 03:43:27 cl Exp $

##
# The IScan class provides all the interface functions for IScan and
#	coordinates the activites of the other classes and modules that compose
#	the IScan application.
#


use strict;
use warnings;

use Bio::IScan::Global;
use Bio::IScan::Insertion_Sequence;
use Bio::IScan::Search_Request;
use Bio::IScan::Search_Response;
use Bio::IScan::Simple_Alignment;
use Bio::IScan::Utilities;

use Bio::IScan::Server::Alignment_Module;
use Bio::IScan::Server::BLAST_Controller;

use Bio::AlignIO;
use Bio::LocatableSeq;
use Bio::Range;
use Bio::Root::Root;
use Bio::SearchIO;


package Bio::IScan::Server::IScan;
	our @ISA = 'Bio::Root::Root';
	
	our $VERSION = '1.0.0';


##
# Creates
# 
# This class follows the object creation
#
# @param -IR_expend_length
#   
# @param -DR_expand_length
#   The length 
# @throw
#
# @see
#
sub new {
	my $class = shift;
	my $this = $class->SUPER::new(@_);
		
	my $parameters = [ 'blast_path', 
					   'db_path', 
					   'db_name', 
					   'filter_path', 
					   'matrix_path', 
					   'alignment_type',
					   'blast_db_genetic_code'
					 ];
	my $optional_parameters = {};
	my $args = Utilities::Process_Arguments($parameters, 
											$optional_parameters,
											@_);

	$this->{'blast_controller'} = 
		Bio::IScan::Server::BLAST_Controller->new(
			{ 'blast_path' => $args->{blast_path},
			  'db_name' => $args->{db_name},
			  'db_path' => $args->{db_path},
			  'filter_path' => $args->{filter_path},
			  'matrix_path' => $args->{matrix_path},
			  'db_genetic_code' => $args->{blast_db_genetic_code} 
			});
	
	my $alignment_module = $args->{'alignment_type'};
	if ($alignment_module !~ /::/) {
		$alignment_module = 'Bio::IScan::Server::' . 
							   $args->{'alignment_type'} . '_Alignment_Module';
	}
	eval "require $alignment_module";
	if ($@) {
		$this->throw(-class => 'IScan::Exception',
					 -text => "Exception loading alignment module.\n" .
							  "\tModule: $alignment_module \n" . "$@",
					 -value => $!);
	}
	$this->{'alignment_module'} = 
		Alignment_Module::New({ 'iscan_instance' => $this });

	$this->{'db_length'} = $this->{'blast_controller'}->db_length();

	return bless $this, $class;
}


##
# Search for ISs with the provided AA query sequences. If all of the provided 
#   query AA sequences cannot be found, no IS is returned. Searching for ISs is
#   terminated after the first AA query sequence.
#
# @param queries
#   Query AA sequences to search for.
#
# @param blast_maximum_expect (default)
#   The maximum expect value an HSP can have and still be considered for
#		processing.
# @param blast_minimum_identities (default)
#   The minimum ratio of identities to AAs that an HSP can have and still be
#		considered for processing.
#
# @param blast_expand_length (default)
#   The distance from the first BLAST hit (upstream or downstream) in which
#		another BLAST hit will be considered part of the same IS.
#
# @param dr_expand_length (default)
#   The distance to the right or left from the right or left IR, respectively,
#		in which to search for DRs.
#
# @param ir_expand_backward (default)
#   The distance to the right from the left-most BLAST hit or the distance to
#		the left from the left-most BLAST hit in which to search for an IR.
#		Esentially, C<ir_expand_forward> expands away from BLAST hits and 
#		C<ir_expand_backward> expands into BLAST hits.
#   NOTE: From the above, "backward" is not absolute, but relative to the BLAST
#		hit.
#
# @param ir_expand_forward (default)
#   The distance to the left from the left-most BLAST hit or the disstance to
#		the right from the right-most BLAST hit in which to search for an IR.
#		Esentially, C<ir_expand_forward> expands away from BLAST hits and 
#		C<ir_expand_backward> expands into BLAST hits.
#   NOTE: From the above, "forward" is not absolute, but relative to the BLAST
#		hit.
#
# @return
#   An instance of Insertion_Sequence satisfying all the specified criteria.
#
# @throw IScan_Error
#   
#   
sub search { # instance
	my $this = shift;
	my $request = shift;

	$this->check_search_arguments($request);

	my @insertion_sequences = ();

	$this->progress("Searching " . $this->{'blast_controller'}->db_filename() . "...\n");

	my $response = Bio::IScan::Search_Response->new(
					{ 'version' => $VERSION, 
					  'search_request' => $request 
					});

	my $queries = $request->queries();
	# Debug
	for my $query (@{ $queries }) {
		$this->_debug("Query:\n", $query->seq());
	}
	
	# Find Open Reading Frames (ORFs)
	my $sequence_hsp_sets = $this->_find_orfs($response);

	for my $hsp_sets (@{ $sequence_hsp_sets }) {
		for my $hsp_set (@{ $hsp_sets }) {
			# In order to deal with 1 or N ORFs we identifiy the first (smallest
			#   coordinate) and last (greatest coordinate).
			my $coding_start = (sort({ 
				$a->start('hit') <=> $b->start('hit')
			} @{ $hsp_set }))[0]->start('hit');
			
			my $coding_end = (sort({
				$b->end('hit') <=> $a->end('hit')
			} @{ $hsp_set }))[0]->end('hit');

			my $coding_strand = $hsp_set->[0]->strand('hit');
			
			my $display_name = $hsp_set->[0]->display_name();
			my $sequence_id = 
				Utilities::Parse_FastA_Header($display_name)->locus();
			$this->_debug("Sequence: ", $sequence_id);
			my $sequence_length = $this->{'blast_controller'}->sequence_length(
								   { 
									   'identifier' => $sequence_id 
								   } );
			$this->_debug("Sequence length: ", $sequence_length);			
			$this->_debug("Hit: $coding_start..$coding_end ($coding_strand)");
			
			# All of the HSPs should have the same strand, so just using the strand
			#   of the first HSP should be safe
			if ($coding_strand == 0) {
				$this->warn("BLAST HSP strand cannot be used to search for a DR or IR:\n" .
							"\tStrand: " . $coding_strand . "\n" .
							"\tSequence: " . $sequence_id . "\n" .
							"\tStart of left-most HSP: " . $coding_start . "\n" .
							"\tEnd of right-most HSP: " . $coding_end . "\n");
				next;
			}

			# Extract from the BLAST database a nucleotide sequence large enough to
			#   accomidate finding IR(s) and DR(s).
			# The coordinates do not need to be adjusted because the first and last
			#	nucleotides of the first and last ORF(s) are not included in the IR
			#	expand out length.
			my $search_start = $coding_start - 
							   ($request->ir_expand_out() + $request->dr_expand_out());
			my $search_end = $coding_end + 
							 ($request->ir_expand_out() + $request->dr_expand_out());
			
			$this->_debug("Search space: $search_start..$search_end");

			# Make sure, based on the BLAST hits that there is at least enough sequence
			#   to search the specified distances from the BLAST hits for IRs
			#   there is at least enough sequence to search for 
			if ($search_start < 1) {
				$this->warn("There is not enough space to the left of the " .
							"left-most BLAST hit to search for a DR or IR:\n" .
							"\tSequence: " . $sequence_id . "\n" .
							"\tStarting coordinate of left-most hit: " . 
								$coding_start . "\n" .
							"\tIR forward expand length: " . 
								$request->ir_expand_out() . "\n" .
							"\tDR expand length: " . $request->dr_expand_out() . "\n");
				next;
			}
			if ($search_end > $sequence_length) {
				$this->warn("There is not enough space to the right of the " .
							"right-most BLAST hit to search for a DR or IR:\n" .
							"\tSequence: " . $sequence_id . "\n" .
							"\tEnding coordinate of right-most hit: " . 
								$coding_end . "\n" .
							"\tIR forward expand length: " . 
								$request->ir_expand_out() . "\n" .
							"\tDR expand length: " . $request->dr_expand_out() . "\n");
				next;
			}

			# Calculate statistics
			if ($request->ir_generate_statistics()) {
				$this->{'alignment_module'}->generate_statistics($sequence_id, 
																 $response);
			}

			# Protect agains exceptions when finding an IS.
			eval {
				
				my $search_range = Bio::Range->new('-start' => $search_start, 
												   '-end' => $search_end,
												   '-strand' => $coding_strand);

				my $search_sequence = $this->_retrieve_locatable_sequence($sequence_id,
																		  $search_range);
				$search_sequence->display_id($display_name);
				# If the hit is on the -1 strand complement the retrieved sequence,
				#	which is always from the +1 strand, and reverse the order so that
				#	all of its elements will be printed 5'-3'
				if ($coding_strand < 0) {
					$search_sequence = $search_sequence->revcom();
					$search_sequence->strand($coding_strand);
				}
				
				my $search_length = $search_sequence->length();
				
				# Build the ORF sequences
				my @orf_sequences = ();
				for my $orf (@{ $hsp_set }) {

					my ($cut_start, $cut_end) = 
						$this->_global_to_string_cut($orf->start('hit'), 
													 $orf->end('hit'),
													 $search_sequence);
													 
					my $seq = $search_sequence->subseq($cut_start, $cut_end);
					
					# Note: the strand of the ORF sequence is the strand of the search
					#	sequence it was cut from; this will be changed if we need to
					#	reverse complement the ORF sequence.
					my $orf_sequence = 
						Bio::LocatableSeq->new('-id' => $search_sequence->display_id(),
											   '-start' => $orf->start('hit'), 
											   '-end' => $orf->end('hit'),
											   '-strand' => $search_sequence->strand(),
											   '-seq' => $seq);
					
					# The ORFs do not all have to be on the same strand, so check the
					#	strand of this ORF and adjust the sequence if necessary.
					if ($orf->strand('hit') != $search_sequence->strand()) {
						$orf_sequence = $orf_sequence->revcom();
					}
					
					push(@orf_sequences, $orf_sequence);
				}

				# Extract and search for the IRs
				my $ir_anchor = Bio::Range->new('-start' => $coding_start, 
												'-end' => $coding_end);
				
				my ($ir_search_sequence_l, $ir_search_sequence_r) = 
					$this->_extract_search_sequences({ 'expand_forward' => $request->ir_expand_out(), 
													   'expand_backward' => $request->ir_expand_in(),
													   'sequence' => $search_sequence, 
													   'anchor' => $ir_anchor,
													   'bound' => $search_range });

				my ($l_ir_alignment, $r_ir_alignment) = 
					$this->_find_irs($response, 
									 $ir_search_sequence_l, $ir_search_sequence_r, 
									 $sequence_id);
				
				# Extract and search for the DRs
				my ($lower_ir_start, $upper_ir_end) = undef;
				if ($search_sequence->strand() < 0) {
					$lower_ir_start = $r_ir_alignment->right_sequence()->start();			
					$upper_ir_end = $l_ir_alignment->left_sequence()->end();
				}
				else {
					$lower_ir_start = $l_ir_alignment->left_sequence()->start();			
					$upper_ir_end = $r_ir_alignment->right_sequence()->end();
				}

				my $dr_anchor = Bio::Range->new('-start' => $lower_ir_start,
												'-end' => $upper_ir_end);

				my ($dr_search_sequence_l, $dr_search_sequence_r) =
					$this->_extract_search_sequences({ 'sequence' => $search_sequence,
													   'anchor' => $dr_anchor,
													   'expand_forward' => $request->dr_expand_out(),
													   'expand_backward' => $request->dr_expand_in(),
													   'bound' => $search_range });

				my $dr_alignment = $this->_find_drs($response,
													$dr_search_sequence_l, 
													$dr_search_sequence_r);

				# Build the IS sequence		
				my ($cut_start, $cut_end) = 
					$this->_global_to_string_cut($lower_ir_start,
												 $upper_ir_end,
												 $search_sequence);
				
				$this->_debug("IS: ", $lower_ir_start, "..", $upper_ir_end, 
							  " (", $cut_start, " -> ", $cut_end, ")");

				my $seq = $search_sequence->subseq($cut_start, $cut_end);
				
				my $is_sequence = Bio::LocatableSeq->new('-id' => $search_sequence->display_id(),
														 '-start' => $lower_ir_start,
														 '-end' => $upper_ir_end,
														 '-strand' => $search_sequence->strand(),												 
														 '-seq' => $seq);

				# Build the IS
				my $is = Bio::IScan::Insertion_Sequence->new(
							{
								'name' => $request->search_name(),
								'sequence' => $is_sequence,
								'orfs' => \@orf_sequences,
								'hsps' => $hsp_set,
								'dr_alignment' => $dr_alignment,
								'l_ir_alignment' => $l_ir_alignment,
								'r_ir_alignment' => $r_ir_alignment 
							});

				$response->add_insertion_sequence($is);
			};
			if ($@) {
				$this->warn("Failure searching for insertion sequence:\n" .
							"\tSequence: " . $sequence_id . "\n" .
							"\tCoding Coordinates: " . 
								$coding_start . ".." . $coding_end . 
							" (" . $coding_strand . ")\n" .
							"\tFailure: " . $@);
			}
		}
	}

	$response->mark_execution_end();

	return $response;
}


sub check_search_arguments { # instance
	my $this = shift;
	my $request = shift;
	
	$this->{'alignment_module'}->check_alignment_arguments($request);
	
	if ((my $queries = $request->queries()) > 0) {
		for my $query (@{ $queries }) {
			if (!$query->isa('Bio::SeqI')) {
				$this->throw('-class' => 'Bio::Root::BadParameter',
							 '-text' => "Query must be an instance of <Bio::SeqI>\n" . 
										"\tQuery: " . $query,
							 '-value' => 'Bio::Root::BadParameter');
			}
			elsif ($query->alphabet() ne 'protein') {
				$this->throw('-class' => 'Bio::Root::BadParameter',
							 '-text' => "Query must be a protein sequence\n" .
										"\tQuery: " . $query->display_id() . "\n" .
										"\tAlphabet: " . $query->alphabet(),
							 '-value' => 'Bio::Root::BadParameter');
			}
		}
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "No insertion sequence queries sepecified",
					 '-value' => 'Bio::Root::BadParameter');
	}

	if ((my $value = $request->duplicate_hsp_overlap()) =~ /^[\d\.]+$/o) {
		if ($value < 0 || $value > 1) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "<duplicate_hsp_overlap> " .
									"must be a faction between 0 and 1\n" .
									"\tValue: " . $value,
						 '-value' => 'Bio::Root::BadParameter');		
		}		
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<duplicate_hsp_overlap> " .
					 "must be a faction between 0 and 1\n" .
					 "\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}
	
	if ((my $value = $request->group_hsps_accross_strands()) !~ /^(0|1)$/o) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<group_hsps_accross_strands> must be 0 or 1\n" .
								"\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}
	
	if ((my $value = $request->blast_minimum_hsp_fraction_identical()) =~ /^[\d\.]+$/o) {
		if ($value < 0 || $value > 1) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "<blast_minimum_hsp_fraction_identical> ".
									"must be a fraction between 0 and 1\n" .
									"\tValue: " . $value,
						 '-value' => 'Bio::Root::BadParameter');
		}	
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<blast_minimum_hsp_fraction_identical> " .
								"must be a fraction between 0 and 1\n" .
								"\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}
	
	if ((my $value = $request->blast_minimum_query_alignment_fraction_identical()) =~ /^[\d\.]+$/o) {
		if ($value < 0 || $value > 1) {		
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "<blast_minimum_query_alignment_fraction_identical>" .
						 " must be a fraction between 0 and 1\n" .
						 "\tValue: " . $value,
						 '-value' => 'Bio::Root::BadParameter');
		}
	}
	else {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "<blast_minimum_query_alignment_fraction_identical>" .
						 " must be a fraction between 0 and 1\n" .
						 "\tValue: " . $value,
						 '-value' => 'Bio::Root::BadParameter');		
	}	
}


##
#
sub _find_orfs { # instance
	my $this = shift;

	my $hsp_sets = [];	# Array of set of HSPs (array) for a sequence
						# [sequence][hsp_set][hsp]
	
	my $response = shift;

	my $request = $response->search_request();
	my @queries = @{ $request->queries() };

	my $sequence_hsps = {};	# Hash of array of queries of array of HSPs
							#	organized by sequence
							# {sequence}[query][hsp]

	$this->progress("Finding ORF(s)...\n");
	my $progressed = 0;
	
	
	# Find BLAST hits
	my $query_count = scalar @queries;
	for (my $current_query = 0; $current_query < $query_count; ++$current_query) {
		my $query = $queries[$current_query];
		
		my $blast_io = $this->{blast_controller}->blast_translate($query);

		while (my $result = $blast_io->next_result()) {
			$this->_debug("Query ", $current_query, " Hit(s): ", $result->num_hits());
			
			while (my $hit = $result->next_hit()) {
				$this->_debug("Hit ", $hit->name(), " Total HSP(s): ", 
							  $hit->num_hsps());
				
				my $hsps = {};	# HSPs for the hit being processed
				
				while (my $hsp = $hit->next_hsp()) {
					# Filter out HSP(s) that do not meet specified criteria.
					if ($hsp->expect() > $request->blast_maximum_expect()) {
						next;
					}
					
					if ($hsp->frac_identical('total') < 
						$request->blast_minimum_hsp_fraction_identical()) {
						next;
					}
					
					if (($hsp->num_identical() / $query->length()) <
						$request->blast_minimum_query_alignment_fraction_identical()) {
						next;
					}
					
					# The HSP display name is not normally set.
					$hsp->display_name($hit->name() . ' ' . $hit->description());
					
					$hsps->{$hsp} = $hsp;
				}
				
				# When all the HSPs for a hit have been processed, store them
				#	for pairing.				
				$this->_debug("Hit ", $hit->name(), " Filtered HSP(s): ", 
							  scalar keys %{ $hsps });

				if (!defined $sequence_hsps->{$hit->name()}) {
					$sequence_hsps->{$hit->name()} = [];
				}

				push(@{ $sequence_hsps->{$hit->name()} }, [ values %{ $hsps } ]);
			}
		}		
	}
	
	if ($query_count > 1) {
		for my $sequence_key (keys %{ $sequence_hsps }) {
			my $hsps = $sequence_hsps->{$sequence_key};
			
			$this->_debug("Grouping ", $sequence_key, 
						  " Queries: ", scalar @{ $hsps });

			my $sequence_hsp_sets = $this->__group_hsps($response, $hsps);
			
			push(@{ $hsp_sets }, $sequence_hsp_sets);
		}
		
	}
	# If there is only one query, it is not necessary to group the HSPs into
	#	sets; each of the HSPs in the one array of HSPs for the one query can be
	#	converted into a array of one element arrays each with a single HSP: the
	#	only HSP for a set of one query HSP sets.
	else {
		for my $sequence_key (keys %{ $sequence_hsps }) {			
			# There should never be more than one element, the query element, in
			#	the array of query hsps for a sequence.
			my $query_hsp_groups = scalar @{ $sequence_hsps->{$sequence_key} };
			if ($query_hsp_groups > 1) {
				$this->throw('-class' => 'Bio::Root::BadParameter',
							 '-text' => "The number of query HSPs groups exceeds " .
										"the number of queries:" .
										"\tQuery HSP groups: " . $query_hsp_groups,
							 '-value' => $query_hsp_groups);
			}
	
			my $sequence_hsp_sets = [];
			for my $hsp (@{ $sequence_hsps->{$sequence_key}->[0] }) {
				push(@{ $sequence_hsp_sets }, [ $hsp ]);
			}
			
			push(@{ $hsp_sets }, $sequence_hsp_sets);
		}
	}
	
	$this->progress("\n");

	return $hsp_sets;
}


##
# Group the HSPs for a sequence into unique sets of HSPs that do not violate any
#	of the grouping contraints and contain ~exactly~ the same number of HSPs as
#	queries.
#
sub __group_hsps { # instance
	my $this = shift;
	my $response = shift;
	my $hit_hsps = shift;
	
	# Build a tree of all the "valid" combinations of HSPs.
	my $root_node = $this->_new_hsp_set_builder($response, $hit_hsps)->build();
	
	# Traverse the tree to determine the sets of HSPs.
	my $hsp_sets = $this->__traverse_hsp_tree($root_node);
	
	return $hsp_sets;
}


##
# Traverse an HSP tree built by the I<HSP set builder>, returning an array of 
#	arrays of HSPs.
#
sub __traverse_hsp_tree { # instance
	my $this = shift;
	my $root_node = shift;
	
	my $hsp_sets = [];	# Array of HSP sets (array), representing the unique
						#	sets of HSPs for a sequence
	
	my $root_children = $root_node->child_node_list();
	for my $child_node (@{ $root_children }) {
		$this->_debug("Set: ", scalar @{ $hsp_sets });
		
		my $hsp_set = [];	# Array of HSPs comprised of a unique set of HSPs
							#	that represent the ORFs of an IS
		
		$this->__traverse_hsp_branch($child_node, $hsp_set);
		
		push(@{ $hsp_sets }, $hsp_set);
	}
	
	return $hsp_sets;
}

sub __traverse_hsp_branch { # instance
	my $this = shift;
	my $node = shift;
	my $hsp_set = shift;
	
	$this->_debug("Adding: ", 
				  $node->hsp()->start('hit'), "..", $node->hsp()->end('hit'));
	
	push(@{ $hsp_set }, $node->hsp());
	
	my $node_children = $node->child_node_list();
	for my $child_node (@{ $node_children }) {
		$this->__traverse_hsp_branch($child_node, $hsp_set);
	}
}


##
# Convert a global cut coordinate (gc) to a string cut coordinate (sc).
#
# This function is strand aware: if the <bound> parameter's <strand> method
#	retuns a values less than zero the function calculates the string
#	coordinates for a string that has been reverse complemented.
#
# Note: this method was added after <_extract_search_sequences>, but duplicates
#	some of the functionality in that method; at some point 
#	<_extract_search_sequences> should call this method.
#
sub _global_to_string_cut { # instance
	my $this = shift;
	my $global_cut_start = shift;
	my $global_cut_end = shift;
	my $bound = shift;
	
	my $string_cut_start = undef;
	my $string_cut_end = undef;

	if ($bound->strand() < 0) {
		$string_cut_start = $bound->end() - $global_cut_end + 1;
		$string_cut_end = $bound->end() - $global_cut_start + 1;
	}
	else {
		$string_cut_start = $global_cut_start - $bound->start() + 1;
		$string_cut_end = $global_cut_end - $bound->start() + 1;
	}
	
	return ($string_cut_start, $string_cut_end);
}


##
# Extract a left and right sequences by "forward" and "backward" from a core
#   sequence.
# NOTE: "forward" and "backward" mean to the left and right, respectively, from
#   the left side of the core and rigth and left, respectively, from the right
#   side of the core.
#
# @param sequence
# @param core
# @param forward
# @param backward
# @param offset
#
# @return
#   An instance of Insertion_Sequence satisfying all the specified criteria.
#
# @throw IScan_Error
#   
#
sub _extract_search_sequences { # instance
	my $this = shift;
	
	my ($args) = @_;
	# Convert arguments to lowercase
	$args->{map(lc($_), keys %{$args})} = values %{$args};
	
	my $sequence = $args->{'sequence'};
	my $anchor = $args->{'anchor'};
	my $bound = $args->{'bound'};

	$this->_debug("Anchor: ", $anchor->start(), "..", $anchor->end());

	my ($start_left, $end_left, $lower_cut_start, $lower_cut_end) = undef;
	my ($start_right, $end_right, $upper_cut_start, $upper_cut_end) = undef;

	if ($sequence->strand() < 0) {
		$start_right = $anchor->start() - $args->{'expand_forward'};
		# The '- 1' is because a backward expansion of -1 means to include the
		#   nucletoide from which the expansion is taking place. So, first we need
		#   to exclude it before performing any math.
		$end_right = $anchor->start() + $args->{'expand_backward'} - 1;

		# '+ 1' because 'subseq' is inclusive
		$upper_cut_start = $bound->end() - $end_right + 1;
		$upper_cut_end = $bound->end() - $start_right + 1;
					
		# The '+ 1' is because a backward expansion of -1 means to include the
		#   nucletoide from which the expansion is taking place. So, first we need
		#   to exclude it before performing any math.
		$start_left = $anchor->end() + 1 - $args->{'expand_backward'};
		$end_left = $anchor->end() + $args->{'expand_forward'};
		
		# '+ 1' because 'subseq' is inclusive
		$lower_cut_start = $bound->end() - $end_left + 1;
		$lower_cut_end = $bound->end() - $start_left + 1;
	}
	else {
		$start_left = $anchor->start() - $args->{'expand_forward'};
		# The '- 1' is because a backward expansion of -1 means to include the
		#   nucletoide from which the expansion is taking place. So, first we need
		#   to exclude it before performing any math.
		$end_left = $anchor->start() + $args->{'expand_backward'} - 1;
	
		# '+ 1' because 'subseq' is inclusive
		$lower_cut_start = $start_left - $bound->start() + 1;
		$lower_cut_end = $end_left - $bound->start() + 1;
		
		# The '+ 1' is because a backward expansion of -1 means to include the
		#   nucletoide from which the expansion is taking place. So, first we need
		#   to exclude it before performing any math.
		$start_right = $anchor->end() + 1 - $args->{'expand_backward'};
		$end_right = $anchor->end() + $args->{'expand_forward'};
		
		# '+ 1' because 'subseq' is inclusive
		$upper_cut_start = $start_right - $bound->start() + 1;
		$upper_cut_end = $end_right - $bound->start() + 1;		
	}

	$this->_debug("L. Expand: (+", 
				 $args->{'expand_forward'}, ", -", $args->{'expand_backward'}, ") ", 
				 $start_left, "..", $end_left, 
				 " (", $lower_cut_start, " -> ", $lower_cut_end, ")");
	
	my $sequence_l = $sequence->subseq($lower_cut_start, $lower_cut_end);
	my $search_sequence_l = Bio::LocatableSeq->new('-id' => "left|$start_left..$end_left",
												   '-seq' => $sequence_l,
												   '-start' => $start_left,
												   '-end' => $end_left,
												   '-strand' => $sequence->strand());

	$this->_debug("R. Expand: (+", 
				 $args->{'expand_forward'}, ", -", $args->{'expand_backward'}, ") ", 
				 $start_right, "..", $end_right, 
				 " (", $upper_cut_start, " -> ", $upper_cut_end, ")");

	my $sequence_r = $sequence->subseq($upper_cut_start, $upper_cut_end);
	my $search_sequence_r = Bio::LocatableSeq->new('-id' => "right|$start_right..$end_right",
												   '-seq' => $sequence_r,
												   '-start' => $start_right,
												   '-end' => $end_right,
												   '-strand' => $sequence->strand());
	
	return ($search_sequence_l, $search_sequence_r);
}


##
# Find the upstream and downstream IRs.
# This implementation finds the IRs by performing a local alignment 
#
# @param search_sequence_l
#   The left sequence to search for an IR in.
# @param search_sequence_r
#   The right sequence to search for an IR in.
# @param args
#   C<IScan> search arguments
#
# @return C<Bio::IScan::Simple_Alignment>
#   Alignment generated by aligning the two search sequences
#
sub _find_irs { # instance
	my $this = shift;
	
	my $response = shift;
	my $search_sequence_l = shift;
	my $search_sequence_r = shift;
	my $sequence_id = shift;
	
	my ($l_is_alignment, $r_is_alignment) = 
		$this->{'alignment_module'}->align_irs($response,
											   $search_sequence_l,
											   $search_sequence_r,
											   $sequence_id);

	return ($l_is_alignment, $r_is_alignment);
}


##
# @param search_sequence1
#   The left sequence to search for an IR in.
# @param search_sequence2
#   The right sequence to search for an IR in.
#
sub _find_drs { # instance
	my $this = shift;
	
	my $response = shift;	
	my $search_sequence_l = shift;
	my $search_sequence_r = shift;

	my $is_alignment = 
		$this->{'alignment_module'}->align_drs($response,
											   $search_sequence_l,
											   $search_sequence_r);

	return $is_alignment;
}



##
#
sub _retrieve_locatable_sequence { # instance
	my $this = shift;

	my $id = shift;
	my $range = shift;
	
	my $nucleotide_sequence = 
		$this->{'blast_controller'}->retrieve_sequence($id, 
													   $range->start(), 
													   $range->end());

	my $sequence = Bio::LocatableSeq->new('-seq' => $nucleotide_sequence,
										  '-start' => $range->start(), 
										  '-end' => $range->end(),
										  '-strand' => +1);
	
	return $sequence;
}


##
#
sub _retrieve_locatable_sequences { # instance
	my $this = shift;
	
	my @ranges = @_;
	
	my @sequences = undef;
	
	for my $range (@ranges) {
		my $nucleotide_sequence = 
		$this->{blast_controller}->retrieve_sequence($range->start(), 
													 $range->end());
		my $sequence = 
			Bio::LocatableSeq->new(-seq => $nucleotide_sequence,
								   -start => $range->start(), 
								   -end => $range->end());
		$sequences[$#sequences] = $sequence;
	}
	
	return @sequences;
}


sub db_name { # instance
	my $this = shift;
	
	return $this->{'blast_controller'}->db_filename();
}


sub db_time { # instance
	my $this = shift;
	
	my $newest_time = undef;

	my @db_files = $this->{'blast_controller'}->db_files();
	for my $db_file (@db_files) {
		# Element '9' of 'stat' is the last modification time (mtime)
		my $file_mtime = (stat($db_file))[9];
		
		if (!defined $newest_time) {
			$newest_time = $file_mtime;
		}
		elsif ($file_mtime > $newest_time) {
			$newest_time = $file_mtime;
		}
	}
	
	return $newest_time;
}


sub db_length { # instance
	my $this = shift;
	my $sequence_id = shift;

	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "BLAST database length is immutable",
					 -value => @_);
	}
	
	return $this->{'blast_controller'}->db_length();
}


sub db_sequence_length { # instance
	my $this = shift;
	my $sequence_id = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "BLAST database sequence length is immutable",
					 -value => @_);
	}
	
	return $this->{'blast_controller'}->sequence_length({ 'identifier' => $sequence_id });
}


sub db_sequence { # instance
	my $this = shift;
	my $sequence_id = shift;

	my $sequence_length = $this->db_sequence_length($sequence_id);
	return $this->{'blast_controller'}->retrieve_seq($sequence_id, 
													 1, $sequence_length);
}


sub progress { # instance
	my $this = shift;
	
	Global::Progress(@_);
}


sub show_progress { # instance
	my $this = shift;
	
	return Global::Show_Progress(@_);
}


sub _debug { # instance
	my $this = shift;
	Global::Debug(__PACKAGE__, @_);
}


##
# Exception class for BLAST_Controller.
#
@IScan_Exception::ISA = 'Bio::Root::Exception';


##
# Factory method for returning a new instance of <HSP_Set_Builder>.
#
sub _new_hsp_set_builder { # instance
	my $this = shift;
	my $response = shift;
	my $hsps = shift;
	
	return Bio::IScan::Server::HSP_Tree_Builder->new($response, $hsps);
}


##
# Factory method for returning a new instance of <HSP_Node>.
#
sub _new_hsp_node { # instance
	my $this = shift;
	my $hsp = shift;
	
	return Bio::IScan::Server::HSP_Node->new($hsp);
}


# package Bio::IScan::Server::IScan;


##
# The <HSP_Tree_Builder> is used internally to build a tree of HSPs where each
#	path through the tree corresponds to a unique set of HSPs.
# Build an N-node tree of HSPs where each unique path from the root node to a
#	terminal node:
#	1.	Does not contain the same set of HSPs as another path; and
#	2.	Contains only as many nodes as queries.
#
package Bio::IScan::Server::HSP_Tree_Builder;

sub new { # constructor
	my $class = shift;
	my $this = {};
	
	$this->{'response'} = shift;
	my $query_hsps = shift;

	$this->{'query_hsps'} = [];	
	$this->{'missing_query'} = 0;
	
	my $query_count = scalar @{ $query_hsps };
	if ($query_count != 
		scalar @{ $this->{'response'}->search_request()->queries() }) {
	
		$this->{'missing_query'} = 1;
	}
	else {
		# All the HSPs must be in order by their start coordinate.
		for (my $i = 0; $i < $query_count; ++$i) {
			my @hsps = @{ $query_hsps->[$i] };
			
			Global::Debug(__PACKAGE__, "Query ", $i, " HSPs: ", scalar @hsps);

			my @sorted_hsps = sort({ 
				$a->start('hit') <=> $b->start('hit') 
			} @hsps);
			
			if (!@sorted_hsps) { 
				$this->{'missing_query'} = 1;
				last;
			}
			
			push(@{ $this->{'query_hsps'} }, \@sorted_hsps);
		}
	}
	
	return bless $this, $class;
}

##
#
sub build { # instance
	my $this = shift;
	
	my $root_node = $this->_new_hsp_node(undef);
	
	if ($this->{'missing_query'}) { 
		return $root_node; 
	}

	my $query = 0;
		
	my $hsps = $this->{'query_hsps'}->[$query];
	my $hsp_count = scalar @{ $hsps };
		
	for (my $i = 0; $i < $hsp_count; ++$i) {
		 $this->_debug("HSP: ", 
					   $hsps->[$i]->start('hit'), "..", 
					   $hsps->[$i]->end('hit'),
					   " (", $hsps->[$i]->strand('hit'), ")\t",					   
					   " (query: ", $query, ")");
		 
		my $child_node = $this->_add_hsp_branch($query + 1, $hsps->[$i]);
		
		if (defined $child_node) {
			$root_node->add_child_node($child_node);
		}
	}
	
	return $root_node;
}

sub _add_hsp_branch { # instance
	my $this = shift;
	my $query = shift;
	my $previous_hsp = shift;
	
	my $current_node = undef;
	
	my $request = $this->search_request();
	my $query_count = $this->query_count();
	
	my $hsps = $this->query_hsps($query);
	my $hsp_count = scalar @{ $hsps };
	for (my $i = 0; $i < $hsp_count; ++$i) {

		if (!$request->group_hsps_accross_strands() && 
			$hsps->[$i]->strand('hit') != $previous_hsp->strand('hit')) {
			
			$this->_debug("Different strands: ", 
						  $hsps->[$i]->start('hit'), "..", $hsps->[$i]->end('hit'),
						  " (query: ", $query, ")");
			
			next;
		}
		
		my $hsp_range = Bio::Range->new('-start' => $hsps->[$i]->start('hit'),
										'-end' => $hsps->[$i]->end('hit'),
										'-strand' => $hsps->[$i]->strand('hit'));
		
		my $previous_range = Bio::Range->new('-start' => $previous_hsp->start('hit'),
											 '-end' => $previous_hsp->end('hit'),
											 '-strand' => $previous_hsp->strand('hit'));
		
		if (defined (my $overlap = $hsp_range->intersection($previous_range))) {
			# The overlap's strand is zero when the two ranges are on different
			#	strands; the HSPs are only considered overlapping if they are
			#	on the same strand.
			if ($overlap->strand() != 0) {
				my $minimum_start = (sort({ 
					$a <=> $b 
				} ( $hsps->[$i]->start('hit'), $previous_hsp->start('hit') )))[0];
				
				my $maximum_end = (sort({
					$b <=> $a
				} ( $hsps->[$i]->end('hit'), $previous_hsp->end('hit') )))[0];
				
				my $overlap_fraction = $overlap->length() / 
									   ($maximum_end - $minimum_start + 1);

				if ($overlap_fraction > $request->duplicate_hsp_overlap()) {				
					$this->_debug("Overlaps: ", 
								  $hsps->[$i]->start('hit'), "..", 
								  $hsps->[$i]->end('hit'),
								  " overlap: ", $overlap_fraction, "\t",
								  " (query: ", $query, ")");
					
					next;
				}
			}
		}
		
		# If the two ORFs are farther apart than the maximum ORF distance (or
		#	the absolute value of the negative ORF distance), meaning they 
		#	overlap, they cannot be paired.
		my $distance_end_start = undef;
		my $distance_start_start = undef;
		my $distance_end_end = undef;
				
		if ($hsps->[$i]->strand('hit') != $previous_hsp->strand('hit')) {
			if ($previous_hsp->strand('hit') > 0) {
				$distance_end_start = $hsps->[$i]->start('hit') - $previous_hsp->end('hit');				
				$distance_start_start = $hsps->[$i]->start('hit') - $previous_hsp->start('hit');
				$distance_end_end = $hsps->[$i]->end('hit') - $previous_hsp->end('hit');
			}
			else {
				$distance_end_start = $previous_hsp->start('hit') - $hsps->[$i]->end('hit');
				$distance_start_start = $previous_hsp->start('hit') - $hsps->[$i]->start('hit');				
				$distance_end_end = $previous_hsp->end('hit') - $hsps->[$i]->end('hit');
			}

			if (abs($distance_end_start) > $request->maximum_orf_distance() &&
				abs($distance_start_start) > $request->maximum_orf_distance() &&
				abs($distance_end_end) > $request->maximum_orf_distance()) {
				
				$this->_debug("Too distant: ", 
							  $hsps->[$i]->start('hit'), "..", $hsps->[$i]->end('hit'),
							  " (", $hsps->[$i]->strand('hit'), ")",
							  " distances: ", $distance_end_start,
							  " ", $distance_start_start, "\t",
							  "(query: ", $query, ")");
				
				next;
			}
		}
		else {
			# The two HSPs are on the "+" strand
			if ($hsps->[$i]->strand('hit') > 0) {
				$distance_end_start = $hsps->[$i]->start('hit') - $previous_hsp->end('hit');
			}
			# The two HSPs are on the "-" strand
			else {
				$distance_end_start = $previous_hsp->start('hit') - $hsps->[$i]->end('hit');
			}
			
			if (abs($distance_end_start) > $request->maximum_orf_distance()) {
																		   
				# The debugging statement below produces an excessive amount of
				#	debugging information, and should only be uncommented for
				#	debugging the tree-building algorithm.
				$this->_debug("Too distant: ", 
							  $hsps->[$i]->start('hit'), "..", $hsps->[$i]->end('hit'),
							  " (", $hsps->[$i]->strand('hit'), ")",
							  " distance: ", $distance_end_start, "\t",
							  "(query: ", $query, ")");
																		   
				next;
			}
		}
																   
		$this->_debug("Join: ", 
					  $hsps->[$i]->start('hit'), "..", 
					  $hsps->[$i]->end('hit'),
					  " (", $hsps->[$i]->strand('hit'), ") ",
					  "distances: ", 
					  (defined $distance_end_start) ? $distance_end_start : "", " ", 
					  (defined $distance_start_start) ? $distance_start_start : "", "\t",
					  (defined $distance_end_end) ? $distance_end_end : "", "\t",
					 "(query: ", $query, ")");
																   
		if ($query < $query_count - 1) {
		   my $child_node = $this->_add_hsp_branch($query + 1, $hsps->[$i]);
		   
		   if (defined $child_node) {
			   if (!defined $current_node) {
				   $current_node = $this->_new_hsp_node($previous_hsp);
			   }					
			   $current_node->add_child_node($child_node);
		   }
		}
		else {
		   if (!defined $current_node) {
			   $current_node = $this->_new_hsp_node($previous_hsp);
		   }
		   
		   my $child_node = $this->_new_hsp_node($hsps->[$i]);
		   $current_node->add_child_node($child_node);
		}
	}

	return $current_node;
}

sub query_count { # instance
	my $this = shift;
	
	return scalar @{ $this->search_request()->queries() };
}

sub query_hsps { # instance
	my $this = shift;
	my $query = shift;
	
	return $this->{'query_hsps'}->[$query];
}

sub search_request { # instance
	my $this = shift;
	
	return $this->{'response'}->search_request();
}

##
# Factory method for returning a new instance of <HSP_Node>.
#
sub _new_hsp_node { # instance
	my $this = shift;
	my $hsp = shift;
	
	return Bio::IScan::Server::HSP_Node->new($hsp);
}

sub _debug { # instance
	my $this = shift;
	Global::Debug(__PACKAGE__, @_);
}


# package Bio::IScan::Server::HSP_Tree_Builder;


##
# The HSP_Node is used internally for constructing a tree of HSPs.
#
# Note: This class is ~not~ thread safe.
#
package Bio::IScan::Server::HSP_Node;

sub new { # constructor
	my $class = shift;
	my $this = {};

	$this->{'parent'} = undef;
	$this->{'hsp'} = shift;
	$this->{'children'} = {};
	
	
	return bless $this, $class;
}


sub parent { # instance
	my $this = shift;
	
	if (@_) {
		$this->{'parent'} = shift;
	}
	
	return $this->{'parent'};
}


sub add_child_node { # instance
	my $this = shift;
	my $node = shift;

	$node->parent($this);
	$this->{'children'}->{$node} = $node;
	
	return $node;
}

sub remove_child_node { # instance
	my $this = shift;
	my $node = shift;
	
	delete $this->{'children'}->{$node};

	return $node;
}

sub child_node_list { # instance
	my $this = shift;
	
	my @children = values %{ $this->{'children'} };
	return \@children;
}


sub hsp { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "The HSP a HSP_Node refers to cannot be changed.",
					 -value => @_);
	}
		
	return $this->{'hsp'};
}


# package Bio::IScan::Server::HSP_Node;


1;

__END__
