# $Id: diprofile_Alignment_Module.pm,v 1.1 2007/03/13 04:24:31 cl Exp $

##
# The <diprofile_Alignment_Module> is an alignment module that uses CLUSTALW to
#	globally align each IR against a specified IR sequence.
#

use strict;
# This module cannot use <warnings>, because each alignment module overrides the
#	<Alignment_Module::New> function defined in the subclass alignment module.

use Bio::IScan::Search_Request;
use Bio::IScan::Search_Response;
use Bio::IScan::Template_Alignment;
use Bio::IScan::Simple_Alignment;

use Bio::IScan::Utilities;

use Bio::IScan::Server::CLUSTALW_Controller;


package Bio::IScan::Server::diprofile_Alignment_Module;
	our @ISA = 'Bio::IScan::Server::Alignment_Module';

# Left and righ IR statistic postixes.
our $ir_statistic_separator = '~';
our $left_ir_statistic_id = 'L';
our $right_ir_statistic_id = 'R';
our $left_ir_statistic_postfix = $ir_statistic_separator . $left_ir_statistic_id;
our $right_ir_statistic_postfix = $ir_statistic_separator . $right_ir_statistic_id;

# Default alignment scoring parameters used by CLUSTALW.
our $iub_match_value = 1.9;
our $iub_mismatch_value = 0.0;
our $clustalw_gap_penalty = 10.0;
our $clustalw_extension_penalty = 0.2;

sub Alignment_Module::New { # factory
	return Bio::IScan::Server::diprofile_Alignment_Module->new(@_);
}


sub new { # constructor
	my $class = shift;
	my $this = $class->SUPER::new(@_);
	
	my $parameters = [];
	my $optional_parameters = {};

	my $arguments = Utilities::Process_Arguments($parameters,
												 $optional_parameters,
												 @_);

	map { $this->{$_} = $arguments->{$_}; } keys %{ $arguments };

	$this->{'alignment_controller'} = Bio::IScan::Server::CLUSTALW_Controller->new(
		{ 
			'path' => Global::Configuration_Parameter('align_path'),
			'name' => Global::Configuration_Parameter('align_name') 
		});
	
	return bless $this, $class;
}


sub _new_alignment { # factory
	my $this = shift;
	return Bio::IScan::Template_Alignment->new(@_);
}


# The diprofile alignment type does not align IRs and DRs in the same manner, 
#	so a different alignment type is used for the DR alignments.
sub _new_dr_alignment { # factory
	my $this = shift;
	return Bio::IScan::Simple_Alignment->new(@_);
}


sub check_alignment_arguments { # instance
	my $this = shift;
	my $request = shift;
	
	$this->SUPER::check_alignment_arguments($request);
	
	if ((my $sum = $request->ir_expand_out() + $request->ir_expand_in()) < 3) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "The sum of <ir_expand_in> and <ir_expand_out> " . 
					 "must be grater than 3\n" .
					 "\tSum: " . $sum,
					 '-value' => 'Bio::Root::BadParameter');		
	}
	
	if ((my $sum = $request->dr_expand_out() + $request->dr_expand_in()) < 3) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "The sum of <dr_expand_in> and <dr_expand_out> must be grater than 3\n" .
					 "\tSum: " . $sum,
					 '-value' => 'Bio::Root::BadParameter');		
	}	

	if ((my $value = $request->alignment_mismatch_value()) !~ /^[\d\-]+$/o) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<alignment_mismatch_value>" .
					 " must be an integer grater than 0\n" .
					 "\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}	
	
	if ((my $value = $request->alignment_gap_extension_value()) !~ /^[\d\-]+$/o) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<alignment_gap_extension_value>" .
					 " must be an integer grater than 0\n" .
					 "\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}
	
	if (defined (my $profile = $request->ir_left_profile())) {		
		if (!$profile->isa('Bio::SeqI')) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Left IR profile must be an instance of <Bio::SeqI>\n" . 
						 "\tQuery: " . $profile,
						 '-value' => 'Bio::Root::BadParameter');
		}
		elsif ($profile->alphabet() ne 'dna') {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Left IR profile must be a DNA sequence\n" .
						 "\tQuery: " . $profile->display_id() . "\n" .
						 "\tAlphabet: " . $profile->alphabet(),
						 '-value' => 'Bio::Root::BadParameter');
		}
	}
	
	if (defined (my $profile = $request->ir_right_profile())) {		
		if (!$profile->isa('Bio::SeqI')) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Right IR profile must be an instance of <Bio::SeqI>\n" . 
						 "\tQuery: " . $profile,
						 '-value' => 'Bio::Root::BadParameter');
		}
		elsif ($profile->alphabet() ne 'dna') {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Right IR profile must be a DNA sequence\n" .
						 "\tQuery: " . $profile->display_id() . "\n" .
						 "\tAlphabet: " . $profile->alphabet(),
						 '-value' => 'Bio::Root::BadParameter');
		}
	}
	
	if (defined (my $profile = $request->ir_left_profile())) {		
		if (!$profile->isa('Bio::SeqI')) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Left IR profile must be an instance of <Bio::SeqI>\n" . 
						 "\tProfile: " . $profile,
						 '-value' => 'Bio::Root::BadParameter');
		}
		elsif ($profile->alphabet() ne 'dna') {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Left IR profile must be a DNA sequence\n" .
						 "\tProfile: " . $profile->display_id() . "\n" .
						 "\tAlphabet: " . $profile->alphabet(),
						 '-value' => 'Bio::Root::BadParameter');
		}
		elsif ($profile->length() > 
			   $request->ir_expand_out() + $request->ir_expand_in()) {
			
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Left IR profile length must be less than or " .
						 "equal to <ir_expand_out> + <ir_expand_in>\n" .
						 "\tProfile: " . $profile->display_id() . "\n" .
						 "\tLength: " . $profile->length() . "\n" .
						 "\t<ir_expand_out> + <ir_expand_in>: " . 
						 ($request->ir_expand_out() + $request->ir_expand_in()),
						 '-value' => 'Bio::Root::BadParameter');			
		}
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "Left IR profile must be defined for diprofile alignment type\n" .
					 "\tProfile: " . $profile,
					 '-value' => 'Bio::Root::BadParameter');		
	}

	if (defined (my $profile = $request->ir_left_profile())) {		
		if (!$profile->isa('Bio::SeqI')) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Right IR profile must be an instance of <Bio::SeqI>\n" . 
						 "\tProfile: " . $profile,
						 '-value' => 'Bio::Root::BadParameter');
		}
		elsif ($profile->alphabet() ne 'dna') {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Right IR profile must be a DNA sequence\n" .
						 "\tProfile: " . $profile->display_id() . "\n" .
						 "\tAlphabet: " . $profile->alphabet(),
						 '-value' => 'Bio::Root::BadParameter');
		}
		elsif ($profile->length() > 
			   $request->ir_expand_out() + $request->ir_expand_in()) {
			
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Right IR profile length must be less than or " .
						 "equal to <ir_expand_out> + <ir_expand_in>\n" .
						 "\tProfile: " . $profile->display_id() . "\n" .
						 "\tLength: " . $profile->length() . "\n" .
						 "\t<ir_expand_out> + <ir_expand_in>: " . 
						 ($request->ir_expand_out() + $request->ir_expand_in()),
						 '-value' => 'Bio::Root::BadParameter');			
		}
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "Right IR profile must be defined for diprofile alignment type\n" .
					 "\tProfile: " . $profile,
					 '-value' => 'Bio::Root::BadParameter');		
	}	
}


sub align_irs { # framework
	my $this = shift;
	
	my $response = shift;
	my $search_sequence_l = shift;
	my $search_sequence_r = shift;
	my $sequence_id = shift;

	my $request = $response->search_request();

	$this->_debug("L. IR: ", 
				  $search_sequence_l->start(), "..", $search_sequence_l->end());
	$this->_debug("R. IR: ", 
				  $search_sequence_r->start(), "..", $search_sequence_r->end());
	
	my $align_sequences_l = [ $search_sequence_l, $request->ir_left_profile() ];
	# For consistency with the other alignment strategies
	my $align_sequences_r = [ $search_sequence_r->revcom(), 
							  $request->ir_right_profile()->revcom() ];
	
	my $alignment_l = $this->align_sequences($align_sequences_l, $response);
	my $alignment_r = $this->align_sequences($align_sequences_r, $response);

	my $ir_l = $alignment_l->get_seq_by_pos(1);
	my $ir_r = $alignment_r->get_seq_by_pos(1);
	
	$this->_debug("L. IR Aln.: ", 
				  $ir_l->start(), "..", $ir_l->end(), "\t", $ir_l->seq());
	$this->_debug("R. IR Aln.: ", 
				  $ir_r->start(), "..", $ir_r->end(), "\t", $ir_r->seq());

	# Calculate alignment statistics if necessary
	if ($request->ir_generate_statistics()) {
		my $left_sequence_id = $sequence_id . $left_ir_statistic_postfix;
		$request->context_parameter('statistics_request', $left_ir_statistic_postfix);		
		my $statistics = $this->generate_statistics($left_sequence_id, $response);
		$request->remove_context_parameter('statistics_request', 0);
		my $percentile = $alignment_l->calculate_percentile($statistics);
		$alignment_l->percentile($percentile);

		my $right_sequence_id = $sequence_id . $right_ir_statistic_postfix;
		$request->context_parameter('statistics_request', $right_ir_statistic_postfix);		
		$statistics = $this->generate_statistics($right_sequence_id, $response);
		$request->remove_context_parameter('statistics_request', 0);		
		$percentile = $alignment_r->calculate_percentile($statistics);
		$alignment_r->percentile($percentile);
	}
	
	return ($alignment_l, $alignment_r);
}


sub align_drs { # framework
	my $this = shift;
	
	my $response = shift;	
	my $search_sequence_l = shift;
	my $search_sequence_r = shift;

	my $request = $response->search_request();
	
	my $search_sequences = [ $search_sequence_l, $search_sequence_r ];

	my $alignment = $this->align_sequences($search_sequences, $response);

	# Convert the default <Template_Alignment> used by the module to a 
	#	<Simple_Alignment>.
	print(STDERR "***Score: ", $alignment->score(), "\n");
	my $dr_alignment = $this->_new_dr_alignment(
		{ 
			'simple_align' => $alignment->wrapped_alignment(),
			'match_score' => $iub_match_value,
			'mismatch_score' => $iub_mismatch_value,
			'gap_score' => abs($request->alignment_gap_extension_value()), 
			'ext_score' => abs($request->alignment_gap_extension_value()) 
		});
	$dr_alignment->score($alignment->score());
	
	
	# The start and end coordinates of the IRs have to be adjusted to genomic
	#   coordinates, because (counter to the BioPerl documentation) the
	#   alignment does not keep track of them.
	my $dr_l = $dr_alignment->get_seq_by_pos(1);
	my $dr_r = $dr_alignment->get_seq_by_pos(2);
	
	$this->_debug("L. DR Aln.: ", 
				  $dr_l->start(), " -> ", $dr_l->end(), "\t", $dr_l->seq());
	$this->_debug("R. DR Aln.: ", 
				  $dr_r->start(), " -> ", $dr_r->end(), "\t", $dr_r->seq());

	return $dr_alignment;	
}


sub align_statistics { # framework
	my $this = shift;

	my $sequence_1 = shift;
	my $sequence_2 = shift;
	my $response = shift;

	my $sequences = undef;

	my $request = $response->search_request();
	my $postfix_id = $request->context_parameter('statistics_id');
	if ($postfix_id eq $left_ir_statistic_id) {
		$sequences = [ $request->ir_left_profile(), $sequence_1 ];		
	}
	elsif ($postfix_id eq $right_ir_statistic_id) {
		$sequences = [ $request->ir_right_profile(), $sequence_2 ];

	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "Invalid <statistics_id>\n" .
								"\t<statistics_id>: " . $postfix_id,
					 '-value' => 'Bio::Root::BadParameter');		
	}

	my $alignment = $this->align_sequences($sequences, $response);
	
	return $alignment;
}


sub align_sequences { # framework
	my $this = shift;

	my $sequences = shift;
	my $response = shift;
	
	my $request = $response->search_request();

	my $alignment = $this->{'alignment_controller'}->align_sequences(
		{ 
			'sequences' => $sequences,
		});
	
	my $clustalw_score = $alignment->score();
	my $is_alignment = $this->_new_alignment(
		{ 'simple_align' => $alignment,
		  'match_score' => $iub_match_value,
		  'mismatch_score' => $iub_mismatch_value,
		  'gap_score' => abs($request->alignment_gap_extension_value()), 
		  'ext_score' => abs($request->alignment_gap_extension_value()) 
		});
	$is_alignment->score($clustalw_score);
	
	
	return $is_alignment;
}


##
# <generate_statistics> is called in two situations:
#	1.	Outside of the alignment module to request statistics generation; or
#	2.	Within the alignment module to request statistics generation or 
#		retrieval from cache.
# The superclass method is overriden to correctly set the <sequence_id>, which 
#	must be different for the alignment of left and right IRs against profiles.
#	Therefore, if the <statistics_request> context parameter is set on the 
#	search request, the <sequence_id> has been correctly set for 2., above. If
#	<statistics_request> has not been set, then statistics are generated, if
#	necessary, for 1., above.
#
sub generate_statistics { # instance
	my $this = shift;
	my $sequence_id = shift;
	my $response = shift;

	my $statistics = undef;

	my $statistics_request = 
		$response->search_request()->context_parameter('statistics_request');
	if (defined $statistics_request) {
		$statistics = $this->SUPER::generate_statistics($sequence_id, 
														$response);
	}
	else {
		my $left_sequence_id = $sequence_id . $left_ir_statistic_postfix;
		$statistics = $this->SUPER::generate_statistics($left_sequence_id, 
														$response);
		
		my $right_sequence_id = $sequence_id . $right_ir_statistic_postfix;
		$statistics = $this->SUPER::generate_statistics($right_sequence_id, 
														$response);
	}
	
	return $statistics;	
}

##
# <_generate_statistics> is overriden from the superclass to remove the left and 
#	right sequence designators, which can not be present when this method is 
#	called.
#
sub _generate_statistics { # instance
	my $this = shift;
	my $response = shift;
	my $postfixed_sequence_id = shift;
	
	my ($sequence_id, $postfix) = 
		($postfixed_sequence_id =~ /^(.+)$ir_statistic_separator(.+)$/o);
	
	$response->search_request()->context_parameter('statistics_id', $postfix);

	my $statistics = $this->SUPER::_generate_statistics($response, $sequence_id);
	$response->search_request()->remove_context_parameter('statistics_id');

	return $statistics;
}


sub _debug { # instance
	my $this = shift;
	Global::Debug(__PACKAGE__, @_);
}


1;

__END__
