# $Id: multiprofile_Alignment_Module.pm,v 1.2 2007/03/13 04:25:32 cl Exp $

##
# The <multiprofile_Alignment_Module> is an alignment module that uses CLUSTALW
#	to globally align each IR against a specified IR sequence.
#

use strict;
# This module cannot use <warnings>, because each alignment module overrides the
#	<Alignment_Module::New> function defined in the subclass alignment module.

use Bio::IScan::Search_Request;
use Bio::IScan::Search_Response;
use Bio::IScan::Simple_Alignment;

use Bio::IScan::Utilities;

use Bio::IScan::Server::CLUSTALW_Controller;


package Bio::IScan::Server::multiprofile_Alignment_Module;
	our @ISA = 'Bio::IScan::Server::Alignment_Module';

# Default alignment scoring parameters used by CLUSTALW.
our $iub_match_value = 1.9;
our $iub_mismatch_value = 0.0;
our $clustalw_gap_penalty = 10.0;
our $clustalw_extension_penalty = 0.2;

sub Alignment_Module::New { # factory
	return Bio::IScan::Server::multiprofile_Alignment_Module->new(@_);
}


sub new { # constructor
	my $class = shift;
	my $this = $class->SUPER::new(@_);
	
	my $parameters = [];
	my $optional_parameters = {};

	my $arguments = Utilities::Process_Arguments($parameters,
												 $optional_parameters,
												 @_);

	map { $this->{$_} = $arguments->{$_}; } keys %{ $arguments };

	$this->{'alignment_controller'} = Bio::IScan::Server::CLUSTALW_Controller->new(
		{ 
			'path' => Global::Configuration_Parameter('align_path'),
			'name' => Global::Configuration_Parameter('align_name') 
		});
	
	return bless $this, $class;
}


sub _new_alignment { # factory
	my $this = shift;
	return Bio::IScan::Simple_Alignment->new(@_);
}


sub check_alignment_arguments { # instance
	my $this = shift;
	my $request = shift;
	
	$this->SUPER::check_alignment_arguments($request);
	
	if ((my $sum = $request->ir_expand_out() + $request->ir_expand_in()) < 3) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "The sum of <ir_expand_in> and <ir_expand_out> " . 
					 "must be grater than 3\n" .
					 "\tSum: " . $sum,
					 '-value' => 'Bio::Root::BadParameter');		
	}
	
	if ((my $sum = $request->dr_expand_out() + $request->dr_expand_in()) < 3) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "The sum of <dr_expand_in> and <dr_expand_out> must be grater than 3\n" .
					 "\tSum: " . $sum,
					 '-value' => 'Bio::Root::BadParameter');		
	}	
	
	if ((my $value = $request->alignment_mismatch_value()) !~ /^[\d\-]+$/o) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<alignment_mismatch_value>" .
					 " must be an integer grater than 0\n" .
					 "\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}	
	
	if ((my $value = $request->alignment_gap_extension_value()) !~ /^[\d\-]+$/o) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<alignment_gap_extension_value>" .
					 " must be an integer grater than 0\n" .
					 "\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}
	
	if (defined (my $profile = $request->ir_left_profile())) {		
		if (!$profile->isa('Bio::SeqI')) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Left IR profile must be an instance of <Bio::SeqI>\n" . 
						 "\tProfile: " . $profile,
						 '-value' => 'Bio::Root::BadParameter');
		}
		elsif ($profile->alphabet() ne 'dna') {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Left IR profile must be a DNA sequence\n" .
						 "\tProfile: " . $profile->display_id() . "\n" .
						 "\tAlphabet: " . $profile->alphabet(),
						 '-value' => 'Bio::Root::BadParameter');
		}
		elsif ($profile->length() > 
			   $request->ir_expand_out() + $request->ir_expand_in()) {

			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "Left IR profile length must be less than or " .
									"equal to <ir_expand_out> + <ir_expand_in>\n" .
									"\tProfile: " . $profile->display_id() . "\n" .
									"\tLength: " . $profile->length() . "\n" .
									"\t<ir_expand_out> + <ir_expand_in>: " . 
									($request->ir_expand_out() + $request->ir_expand_in()),
						 '-value' => 'Bio::Root::BadParameter');			
		}
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "Left IR profile must be defined for multiprofile alignment type\n" .
					 "\tProfile: " . $profile,
					 '-value' => 'Bio::Root::BadParameter');		
	}
}

sub align_irs { # framework
	my $this = shift;
	
	my $response = shift;
	my $search_sequence_l = shift;
	my $search_sequence_r = shift;
	my $sequence_id = shift;

	my $request = $response->search_request();

	my $align_sequence_l = $search_sequence_l;
	my $align_sequence_r = $search_sequence_r->revcom();
	my $template = $response->search_request()->ir_left_profile();
	
	my $align_sequences = [ $align_sequence_l, $align_sequence_r, $template ];

	$this->_debug("L. IR: ", 
				  $align_sequence_l->start(), "..", $align_sequence_l->end());
	$this->_debug("R. IR: ", 
				  $align_sequence_r->start(), "..", $align_sequence_r->end());
	
	my $alignment = $this->align_sequences($align_sequences, $response);

	my $ir_l = $alignment->get_seq_by_pos(1);
	my $ir_r = $alignment->get_seq_by_pos(2);
	
	$this->_debug("L. IR Aln.: ", 
				  $ir_l->start(), "..", $ir_l->end(), "\t", $ir_l->seq());
	$this->_debug("R. IR Aln.: ", 
				  $ir_r->start(), "..", $ir_r->end(), "\t", $ir_r->seq());
		
	$this->_debug("L. IR Aln. (Adj.): ", $ir_l->start(), "..", $ir_l->end());
	$this->_debug("R. IR Aln. (Adj.): ", $ir_r->start(), "..", $ir_r->end());

	# Calculate alignment statistics if necessary
	if ($request->ir_generate_statistics()) {
		my $statistics = $response->ir_statistics($sequence_id);

		my $percentile = $alignment->calculate_percentile($statistics);
		$alignment->percentile($percentile);
	}
	
	return ($alignment, $alignment);
}


sub align_drs { # framework
	my $this = shift;

	my $response = shift;	
	my $search_sequence_l = shift;
	my $search_sequence_r = shift;
	
	my $search_sequences = [ $search_sequence_l, $search_sequence_r ];

	my $request = $response->search_request();
	
	my $alignment = $this->align_sequences($search_sequences, $response);
	
	# The start and end coordinates of the IRs have to be adjusted to genomic
	#   coordinates, because (counter to the BioPerl documentation) the
	#   alignment does not keep track of them.
	my $dr_l = $alignment->get_seq_by_pos(1);
	my $dr_r = $alignment->get_seq_by_pos(2);
	
	$this->_debug("L. DR Aln.: ", 
				  $dr_l->start(), " -> ", $dr_l->end(), "\t", $dr_l->seq());
	$this->_debug("R. DR Aln.: ", 
				  $dr_r->start(), " -> ", $dr_r->end(), "\t", $dr_r->seq());
		
	$this->_debug("L. DR Aln. (Adj.): ", 
				  $dr_l->start(), " -> ", $dr_l->end(), "\t", $dr_l->seq());
	$this->_debug("R. DR Aln. (Adj.): ", 
				  $dr_r->start(), " -> ", $dr_r->end(), "\t", $dr_r->seq());

	return $alignment;	
}


sub align_statistics { # framework
	my $this = shift;
	my $sequence_1 = shift;
	my $sequence_2 = shift;
	my $response = shift;
	
	my $template = $response->search_request()->ir_left_profile();
	my $sequences = [ $sequence_1, $sequence_2, $template];

	my $alignment = $this->align_sequences($sequences, $response);
	
	return $alignment;
}


sub align_sequences { # framework
	my $this = shift;
	my $sequences = shift;
	my $response = shift;
	
	my $request = $response->search_request();
	
	my $alignment = $this->{'alignment_controller'}->align_sequences(
		{ 
			'sequences' => $sequences,
		});
	
	my $clustalw_score = $alignment->score();
	my $is_alignment = $this->_new_alignment(
		{ 'simple_align' => $alignment,
		  'match_score' => $iub_match_value,
		  'mismatch_score' => $iub_mismatch_value,
		  'gap_score' => abs($request->alignment_gap_extension_value()), 
		  'ext_score' => abs($request->alignment_gap_extension_value()) 
		});
	$is_alignment->score($clustalw_score);
	
	
	return $is_alignment;
}


sub _debug { # instance
	my $this = shift;
	Global::Debug(__PACKAGE__, @_);
}


1;

__END__
